use beteran_common_rust as bcr;
use beteran_protobuf_rust as bpr;
use prost::Message;

///
pub struct Service {
  connection_broker: nats::asynk::Connection,
  queue_broker: String,
}

impl std::fmt::Debug for Service {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("beteran-frontend-server-edge::game::service::Service")
      .finish()
  }
}

impl Service {
  ///
  pub fn new(connection_broker: nats::asynk::Connection, queue_broker: String) -> Service {
    Service {
      connection_broker,
      queue_broker,
    }
  }

  pub async fn subscribe(&self) -> std::result::Result<(), std::boxed::Box<dyn std::error::Error>> {
    futures::try_join!(self.list_games(), self.get_game_url(),).map(|_| ())
  }

  fn get_client_in_header(
    &self,
    message: &nats::asynk::Message,
  ) -> Result<bpr::models::core::network::Client, bcr::error::rpc::Error> {
    match &message.headers {
      Some(headers) => {
        let client = match headers.get(bpr::c2se::core::network::HEADER_CLIENT) {
          Some(c) => {
            let msg = base64::decode(c).map_err(|e| {
              bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
                message: format!("invalid header: {}", e),
              })
            })?;
            bpr::models::core::network::Client::decode(msg.as_slice()).map_err(|e| {
              bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
                message: format!("invalid header: {}", e),
              })
            })?
          }
          None => {
            return Err(bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
              message: "invalid client information".to_string(),
            }));
          }
        };

        Ok(client)
      }
      None => Err(bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
        message: "invalid header".to_string(),
      })),
    }
  }

  async fn get_member_session_with_renewal(
    &self,
    client: &bpr::models::core::network::Client,
  ) -> Result<Option<bpr::models::member_session::MemberSession>, bcr::error::rpc::Error> {
    let ss_get_member_session_req = bpr::ss::member_session::GetMemberSessionWithRenewalRequest {
      client: Some(client.clone()),
      request: Some(bpr::ss::member_session::get_member_session_with_renewal_request::Request {}),
    };

    let ss_get_member_session_res_msg = self
      .connection_broker
      .request(
        bpr::ss::member_session::SUBJECT_GET_MEMBER_SESSION_WITH_RENEWAL,
        ss_get_member_session_req.encode_to_vec(),
      )
      .await
      .map_err(|e| {
        bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
          code: bpr::protobuf::rpc::Error::SERVER_00,
          message: format!("server {}", e),
          data: None,
        })
      })?;

    let ss_get_member_session_res =
      bpr::ss::member_session::GetMemberSessionWithRenewalResponse::decode(
        ss_get_member_session_res_msg.data.as_slice(),
      )
      .map_err(|e| {
        bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
          code: bpr::protobuf::rpc::Error::SERVER_00,
          message: format!("server {}", e),
          data: None,
        })
      })?;

    if let Some(e) = ss_get_member_session_res.error {
      return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
        code: bpr::protobuf::rpc::Error::SERVER_00,
        message: format!("server {}", e),
        data: None,
      }));
    }

    if let Some(r) = ss_get_member_session_res.result {
      return Ok(r.member_session);
    }

    Ok(None)
  }

  async fn list_games(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::api::game::SUBJECT_LIST_GAMES,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;

        let req =
          bpr::c2se::api::game::ListGamesRequest::decode(message.data.as_slice()).map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let ss_list_games_req = bpr::ss::api::game::ListGamesRequest {
          client: Some(client),
          request: Some(bpr::ss::api::game::list_games_request::Request {
            pagination: req.pagination,
            sorts: req.sorts,
            search: match req.search {
              Some(s) => Some(bpr::ss::api::game::list_games_request::request::Search {
                parent_id: s.parent_id,
                key_like: s.key_like,
                category_like: s.category_like,
                platform_like: s.platform_like,
                game_type_like: s.game_type_like,
              }),
              None => None,
            },
          }),
        };

        let ss_list_games_res_msg = self
          .connection_broker
          .request(
            bpr::ss::api::game::SUBJECT_LIST_GAMES,
            ss_list_games_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let ss_list_games_res =
          bpr::ss::api::game::ListGamesResponse::decode(ss_list_games_res_msg.data.as_slice())
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;

        if let Some(e) = ss_list_games_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          }));
        }

        if let Some(r) = ss_list_games_res.result {
          message
            .respond(
              bpr::c2se::api::game::ListGamesResponse {
                error: None,
                result: Some(bpr::c2se::api::game::list_games_response::Result { games: r.games }),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::api::game::ListGamesResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn get_game_url(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::api::game::SUBJECT_GET_GAME_URL,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;
        let member_session = match self.get_member_session_with_renewal(&client).await? {
          Some(ms) => ms,
          None => {
            return Err(bcr::error::rpc::Error::InvalidRequest(
              bcr::error::rpc::InvalidRequest {
                message: "Unauthorized".to_string(),
              },
            ));
          }
        };

        let req = bpr::c2se::api::game::GetGameUrlRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let ss_get_game_url_req = bpr::ss::api::game::GetGameUrlRequest {
          client: Some(client),
          request: Some(bpr::ss::api::game::get_game_url_request::Request {
            vendor_key: req.vendor_key,
            game_key: req.game_key,
            member_id: member_session.member_id,
          }),
        };

        let ss_get_game_url_res_msg = self
          .connection_broker
          .request(
            bpr::ss::api::game::SUBJECT_GET_GAME_URL,
            ss_get_game_url_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let ss_get_game_url_res =
          bpr::ss::api::game::GetGameUrlResponse::decode(ss_get_game_url_res_msg.data.as_slice())
            .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        if let Some(e) = ss_get_game_url_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          }));
        }

        if let Some(r) = ss_get_game_url_res.result {
          message
            .respond(
              bpr::c2se::api::game::GetGameUrlResponse {
                error: None,
                result: Some(bpr::c2se::api::game::get_game_url_response::Result {
                  user_id: r.user_id,
                  balance: r.balance,
                  url: r.url,
                }),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::api::game::GetGameUrlResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }
}
